package uk.co.hcsoftware.jdep;

import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Enumeration;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.util.logging.Logger;
import java.util.logging.Level;


/**
 * Title: ClassFinder
 * Copyright:    Copyright (c) 20-Feb-2006
 *
 * @author N Spurling
 */
public class ClassFinder {
    public static final Logger log = Logger.getLogger(ClassFinder.class.getName());
    private List<File> elements;

    public ClassFinder(String classPath) {
        String[] elements = classPath.split(File.pathSeparator);//first split the classpath into tokens
        this.elements=new ArrayList<File>(elements.length);
        for (String element:elements) {
            File f = new File(element);
            if(f.exists()){
                this.elements.add(f);
            }
            else{
               log.info("ignoring classpath element="+element+" as it doesn't exist");
            }
        }
        log.fine("classpath elements: "+this.elements);
    }
    public List<ClassFile> findClassFiles() throws IOException {
        List<ClassFile> classes = new ArrayList<ClassFile>();
        for(File f:elements){
            if(!f.exists()){
                log.fine("ignoring classpath element="+f+" as doesn't exist");
            }
            else if(f.isDirectory()){
                addClasses(f,"", classes);
            }
            else if(f.getName().endsWith(".jar") || f.getName().endsWith(".zip")){ //jar file
                populateFromJar(f, classes);
            }
            else{
                log.fine("ignoring classpath element="+f+" as format not recognized (not a directory, not a jar/zip)");
            }
        }

        return classes;
    }
    private List<ZipFile> openZips=new ArrayList<ZipFile>();
    private void populateFromJar(File f, List<ClassFile> addTo) throws IOException {
        ZipFile zf = new ZipFile(f,ZipFile.OPEN_READ);
        Enumeration<? extends ZipEntry> entries = zf.entries();
        while (entries.hasMoreElements()) {
            ZipEntry ze = entries.nextElement();
            if (ze.getName().endsWith(".class")) {
                addTo.add(new ZipEntryClassFile(ze,zf));
            }
        }
        openZips.add(zf);
    }
    public void close(){
        for(ZipFile zf:openZips){
            try {
                zf.close();
            } catch (IOException e) {
                log.warning("error closign zipfile="+zf+": "+ e);
            }
        }
    }

    private void addClasses(File inDirectory, String packageName, List<ClassFile> addTo) {
        for(File child:inDirectory.listFiles()){
            if(child.isDirectory()){
                final String newPackageName;
                if(packageName.length()>0){
                     newPackageName= packageName+"."+child.getName();
                }
                else newPackageName=child.getName();
                addClasses(child,newPackageName,addTo);
            }
            else if(child.getName().endsWith(".class")){//found a class file!
                final FileClassFile fc = new FileClassFile(packageName,child);
                addTo.add(fc);
                log.finer("found class="+fc);
            }
            else{
                if(log.isLoggable(Level.FINER)){
                    log.finer("ignoring file="+child);
                }
            }
        }
    }
    public interface StreamHandler{
        public void handle(InputStream is) throws IOException, BadClassFileException;
    }
    public interface ClassFile{
        String getClassname();
        public void read(StreamHandler sh) throws IOException, BadClassFileException;
        public ReferenceFinder getReferenceFinder();
    }
    private abstract class AbstractClassFile implements ClassFile{
        private ReferenceFinder referenceFinder;

        public ReferenceFinder getReferenceFinder() {
            if (referenceFinder==null) {
                referenceFinder = new ReferenceFinder(this);
            }
            return referenceFinder;
        }
    }
    private class ZipEntryClassFile extends AbstractClassFile{
        private ZipEntry ze;
        private final ZipFile zf;
        private String classname;

        public ZipEntryClassFile(ZipEntry ze, ZipFile zf) {
            this.ze = ze;
            this.zf = zf;
            // /uk/co/hcsoftware/jdep/Main.class
            String name = ze.getName();
            name=name.replaceAll("/",".");
            if(name.startsWith("/")){
                name=name.substring(1);
            }
            this.classname=name.substring(0,name.length()-".class".length());
        }
         public void read(StreamHandler sh) throws IOException, BadClassFileException {
            InputStream is = zf.getInputStream(ze);
            try {
                sh.handle(is);
            } finally {
                is.close();
            }
        }
        public String getClassname() {
            return classname;
        }

        public String toString() {
            return "name="+classname;
        }
    }
    private class FileClassFile extends AbstractClassFile{
        private File f;
        private String classname;
        public FileClassFile(String packageName, File f) {
            this.f = f;
            final String name = f.getName();
            this.classname=packageName+'.'+name.substring(0,name.length()-".class".length());
        }

        public void read(StreamHandler sh) throws IOException, BadClassFileException {
            log.finest("read classfile="+f+" classname="+classname);
            InputStream is = new BufferedInputStream(new FileInputStream(f));
            try {
                sh.handle(is);
            } finally {
                is.close();
            }
        }
        public String getClassname() {
            return classname;
        }

        public String toString() {
            return "name="+classname+" f="+f;
        }
    }
}
