package uk.co.hcsoftware.jdep;


import ns.TaskTimer;
import uk.co.hcsoftware.yago.ArgReader;
import uk.co.hcsoftware.yago.Argument;
import uk.co.hcsoftware.yago.Option;
import uk.co.hcsoftware.yago.parsers.GnuParser;
import uk.co.hcsoftware.yago.util.DefaultArgReaders;

import java.io.IOException;
import java.io.File;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * give it a classpath arg and some inputclasses and it prints all classes on which they depend to stdout
 *
 * Title: Main
 * Copyright:    Copyright (c) 20-Feb-2006
 *
 * @author N Spurling
 */
public class Main {
    public static final Logger log = Logger.getLogger(Main.class.getName());

    private static ArgReader<String> stringReader = DefaultArgReaders.createStringReader();

    public static final Option<String> classpath = new Option<String>("classpath",true,'c',"classpath",stringReader,"classpath value (same syntax as -cp arg to java "+File.pathSeparator+"-delimited)");
    public static final Option<String> ignorePrefix = new Option<String>("ignore",false,'i',"ignore",stringReader,"ignore any classes whose fully-qualified name starts with one of these values ("+File.pathSeparator+"-separated list)");
    public static final Option<String> onlyPrefix = new Option<String>("only",false,'o',"only",stringReader,"only include classes whose fully-qualified name starts with one of these values ("+File.pathSeparator+"-separated list)");

    public static final Option noRecurse = new Option<Boolean>("norecurse",false,'n',"norecurse",null,"just resolve the input classfile, don't attempt to resolve the classes it references" );
    public static final Option quiet = new Option<Boolean>("quiet",false,'q',"quiet",null,"just print out resolved classes (or error message if bad invocation) -no extra info" );
    public static final Option sort = new Option<Boolean>("sort",false,'s',"sort",null,"sort output classnames lexically" );

    public static final Option warnDynamic = new Option<Boolean>("warndynamic",false,'w',"warndynamic",null,"look for invocations of Class.forName(..) and ClassLoader.loadClass(..) and warn about them [nothing is printed in --quiet mode]" );

    private static Option[] opts = {classpath,ignorePrefix,onlyPrefix,noRecurse,quiet,sort,warnDynamic};
    private static Argument[] arguments = new Argument[]{};
    GnuParser gp = new GnuParser(opts,arguments);
    private List<String> prefixesList;
    private static final String LONG_DESCRIPTION = "prints out the fully-qualified names of all the classes referenced by the input class(es)";
    private static final String INVOCATION = "jdep options inputclass1 [inputclass2.. inputclassn]";
    private ArrayList<String> onlyPrefixList;

    public static void main(String args[]){

        new Main().run(args);
    }

    HashMap<String,ClassFinder.ClassFile> classFiles ;
    private void run(String[] args) {
        try{
            gp.parseOrDie(INVOCATION,args,LONG_DESCRIPTION);

            List<String> argList = gp.getExtraArgs();
            if(argList.isEmpty()){
                System.out.println(gp.getUsage(INVOCATION,LONG_DESCRIPTION));
                System.exit(1);
            }
            log.info("jdep version 0.1");
            TaskTimer tt1=new TaskTimer();
            final String cp = classpath.getArgValue();
            //first we index the classfiles
            log.fine("scanning classpath="+cp+"..");
            TaskTimer tt = new TaskTimer();
            ClassFinder cf = new ClassFinder(cp);
            Collection<String> deps;
            try {
                List<ClassFinder.ClassFile> l = cf.findClassFiles();
                classFiles = new HashMap<String, ClassFinder.ClassFile>();
                for(ClassFinder.ClassFile classfile:l){
                    classFiles.put(classfile.getClassname(),classfile);
                }
                log.fine("completed scan "+l.size()+" classfiles, "+tt.stopString());

                deps = new HashSet<String>();
                if(ignorePrefix.isSet()){
                    String ignore = ignorePrefix.getArgValue();
                    String[] prefixes = ignore.split(File.pathSeparator);
                    prefixesList = new ArrayList<String>(prefixes.length);
                    for(String p:prefixes){
                        String s=p.trim();
                        if(s.length()>0) prefixesList.add(s);
                    }
                }
                if(onlyPrefix.isSet()){
                    String ignore = onlyPrefix.getArgValue();
                    String[] prefixes = ignore.split(File.pathSeparator);
                    onlyPrefixList = new ArrayList<String>(prefixes.length);
                    for(String p:prefixes){
                        String s=p.trim();
                        if(s.length()>0) onlyPrefixList.add(s);
                    }
                }
                tt=new TaskTimer();
                for (String mainClass:argList) {
                    findDeps(mainClass,deps, !noRecurse.isSet());
                }
                if(sort.isSet()){
                    deps = new ArrayList<String>(deps);
                    Collections.sort((List<String>)deps);
                }
                printDeps(deps);
                if (!quiet.isSet()) {
                    System.out.println(deps.size()+" dependencies from inputclasses: "+argList+" ("+l.size()+" classes in classpath)");
                    System.out.println(tt.stopString()+" to find dependencies");
                    System.out.println(tt1.stopString()+" in total");
                    if(warnDynamic.isSet()){
                        StringBuilder warning=new StringBuilder();
                        int dyn=0;
                        for(String cls:deps){
                            ClassFinder.ClassFile cf1 = classFiles.get(cls);
                            if(cf1!=null && cf1.getReferenceFinder().usesDynamicClassloading()){
                                dyn++;
                                warning.append(cf1.getClassname());
                                warning.append("\n");
                            }
                        }
                        if(dyn>0){
                            warning.insert(0,"WARNING: at least "+dyn+(dyn==1?" class uses":" classes use")+" dynamic classloading:\n");
                            System.out.println(warning.toString());
                        }
                    }
                }
            } finally {
               cf.close();
            }


        } catch (Exception e) {
            log.log(Level.SEVERE,"fatal exception",e);
            e.printStackTrace();
        }
    }

    private void printDeps(Collection<String> deps) {
        for(String s:deps){
            System.out.println(s);
        }

    }
    private void findDeps(String className, Collection<String> classNamesResolved, boolean recurse) throws IOException, BadClassFileException {
        ClassFinder.ClassFile cf = classFiles.get(className);
        if(cf==null){
            log.info("ignoring references in class="+className+" as classfile not found");
            return;
        }
        log.finer("resolving class="+className+" recurse="+recurse);
        //load the classbytes and look for references to other classes
        final ReferenceFinder ri = cf.getReferenceFinder();
        if(prefixesList!=null && !prefixesList.isEmpty()){
            ri.setPrefixesToIgnore(prefixesList);
        }
         if(onlyPrefixList!=null && !onlyPrefixList.isEmpty()){
            ri.setPrefixesToInclude(onlyPrefixList);
        }
        Collection<String> references = ri.findReferencedClassnames();

        log.finer("found "+references.size()+" refs");
        log.finest("class="+className+"  refs="+references);

        for(String classname:references){
            if (!classNamesResolved.contains(classname)) { //ignore if we've already seen this classname
                classNamesResolved.add(classname);
                if (recurse) {
                    if(!classFiles.containsKey(classname)){
                        log.info("ignoring references in class="+classname+" from class="+className+" as classfile not found");
                    }
                    else findDeps(classname,classNamesResolved, recurse);
                }
            }
        }
    }
}
