package uk.co.hcsoftware.jdep;

import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.*;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


/**
 * Title: ReferenceFinder
 * Copyright:    Copyright (c) 21-Feb-2006
 *
 * @author N Spurling
 */
public class ReferenceFinder implements ClassFinder.StreamHandler{
    public static final Logger log = Logger.getLogger(ReferenceFinder.class.getName());
    Set<String> classnames;
    private List<String> prefixesToIgnore;
    private ClassFinder.ClassFile cf;
    private final Pattern classDescriptorPattern = Pattern.compile("L.*?;");
    private ArrayList<String> onlyPrefixList;

    public ReferenceFinder(ClassFinder.ClassFile cf) {
        this.cf = cf;
    }

    public void setPrefixesToIgnore(List<String> prefixesToIgnore) {
        this.prefixesToIgnore = prefixesToIgnore;
    }

    List<MethodInfo> methods;
    List<FieldInfo> fields;
    public void handle(InputStream is) throws BadClassFileException {
        try {
            DataInputStream dis = new DataInputStream(is);
            log.finer("parsing file="+cf);

            if(dis.readInt() != 0xcafebabe){//check magic number
                throw new BadClassFileException("not a valid classfile (bad magic number)! ");//+Byte.toString(header[i]));
            }

            skip(dis,4);// classversion
            //from here on we just assume the classfile is kosher, and don't check for read erros
            //read the constantpool
            int constantPoolCount=dis.readUnsignedShort();
            log.finer("constantPoolCount="+constantPoolCount);
            readConstantPool(constantPoolCount, dis);

            skip(dis,2+2+2);//accessflags,thisclassindex,superIndex
            int interfacesCount=dis.readUnsignedShort();
            List<ClassInfo> interfaces=new ArrayList<ClassInfo>(interfacesCount);
            for(int i=0;i<interfacesCount;i++){
                interfaces.add((ClassInfo) constantPool[dis.readUnsignedShort()]);
            }
            //add any classes referenced in fields
            int fieldsCount=dis.readUnsignedShort();
            fields = new ArrayList<FieldInfo>(fieldsCount);
            for(int i=0;i<fieldsCount;i++){
                final FieldInfo fi = readFieldInfo(dis);
                fields.add(fi);
                addClasses(fi.getDescriptor(),classnames);
            }
            //add any classes referenced in methods
            int methodsCount=dis.readUnsignedShort();
            methods=new ArrayList<MethodInfo>(methodsCount);
            for(int i=0;i<methodsCount;i++){
                final MethodInfo methodInfo = readMethodInfo(dis);
                methods.add(methodInfo);
                addClasses(methodInfo.getDescriptor(),classnames);
            }
            //add any classinfo objects
            for(Object o:constantPool){
                if(o instanceof ClassInfo){
                    ClassInfo ci = (ClassInfo) o;
                    String fullyQualifiedClassname = ci.getClassName();
                    if (fullyQualifiedClassname != null && includeClass(fullyQualifiedClassname)) {
                        classnames.add(fullyQualifiedClassname);
                    }
                }
                else if(o instanceof NameAndType){
                    NameAndType nat = (NameAndType) o;
                    for(String fullyQualifiedClassname:nat.getReferencedClassnames()){
                        if (fullyQualifiedClassname != null && includeClass(fullyQualifiedClassname)) {
                            classnames.add(fullyQualifiedClassname);
                        }
                    }
                }
            }
        } catch (IOException e) {
            throw new BadClassFileException("IOError reading file="+cf,e);
        }
    }

    /**
     * look for classnames in methFieldDescriptor and if any found add them to classnames set
     *
     * @param methFieldDescriptor format is as given in http://java.sun.com/docs/books/vmspec/2nd-edition/html/ClassFile.doc.html section 4.3
     * @param classnames
     */
    private void addClasses(String methFieldDescriptor, Set<String> classnames) {
        final List<String> fullyQualifiedClassnames = findClassnames(methFieldDescriptor);
        for (String fullyQualifiedClassname:fullyQualifiedClassnames) {
            if (includeClass(fullyQualifiedClassname)) {
                classnames.add(fullyQualifiedClassname);
            }
        }
    }

    private void skip(DataInputStream is, int bytes) throws IOException {
        int skipped=is.skipBytes(bytes);
        while(skipped<bytes){
            is.readUnsignedByte();
            skipped++;
        }
    }


    private boolean includeClass(String fullyQualifiedClassname){
        if(fullyQualifiedClassname.equals(cf.getClassname())){
            return false;//remove self
        }
        if (prefixesToIgnore!=null) {
            for (String prefix:prefixesToIgnore) {
                if(fullyQualifiedClassname.startsWith(prefix)){
                    return false;
                }
            }
        }
        if (onlyPrefixList!=null) {
            for (String prefix:onlyPrefixList) {
                if(!fullyQualifiedClassname.startsWith(prefix)){
                    return false;
                }
            }
        }
        return true;
    }


    /**
     * see classfile spec, classes are written in classfiles with / delimiters instead of . 'for historical reasons'
     * spec refers to his as 'internal' format
     * @param internalForm
     * @return
     */
    private String decodeInternalName(String internalForm) {
        String ret = internalForm.replaceAll("/",".");
        log.finest("decoded classname="+internalForm+" to ret="+ret);
        return ret;
    }

    public static final byte   CONSTANT_Class =	7
    ,CONSTANT_Fieldref =	9
    ,CONSTANT_Methodref =	10
    ,CONSTANT_InterfaceMethodref =	11
    ,CONSTANT_String 	=8
    ,CONSTANT_Integer =	3
    ,CONSTANT_Float 	=4
    ,CONSTANT_Long =	5
    ,CONSTANT_Double 	=6
    ,CONSTANT_NameAndType 	=12
    ,CONSTANT_Utf8 =	1  ;

    public Collection<String> findReferencedClassnames() throws IOException, BadClassFileException {
        classnames=new HashSet<String>();
        cf.read(this);
        return classnames;
    }
    public boolean usesDynamicClassloading(){
        for(MethRef mr:methodRefs){
            String cls = mr.getClassInfo().getClassName();
            if("java.lang.Class".equals(cls) && "forName".equals(mr.getNameAndType().getName())){
                log.finest("cf="+cf+" uses dynaclassloading meth="+mr);
                return true;
            }
            else if("java.lang.ClassLoader".equals(cls) && "loadClass".equals(mr.getNameAndType().getName())){
                log.finest("cf="+cf+" uses dynaclassloading meth="+mr);
                                return true;
            }
        }
        return false;
    }


    /**
     * extract any fully-qualififed class names from this method or field descriptor.
     * See
     * http://java.sun.com/docs/books/vmspec/2nd-edition/html/ClassFile.doc.html#1169
     * for details of how these are encoded
     * @param encodedDescriptor
     * @return
     */
    private List<String> findClassnames(String encodedDescriptor) {
        List<String> classnames = new ArrayList<String>();
        Matcher m = classDescriptorPattern.matcher(encodedDescriptor);
        while(m.find()){
            String internalForm = m.group();
            internalForm=internalForm.substring(1,internalForm.length()-1);
            classnames.add(decodeInternalName(internalForm));
        }

        log.finest("found classnames="+classnames+" in descriptor="+encodedDescriptor);
        return classnames;
    }
    private List<MethRef> methodRefs;
    private void readConstantPool(int constantPoolCount, DataInputStream is) throws IOException, BadClassFileException {
        constantPool=new Object[constantPoolCount];
        methodRefs = new ArrayList<MethRef>();
        for(int cpi=1;cpi<constantPoolCount;cpi++){
            int tag = is.readUnsignedByte();
            switch(tag){
                case CONSTANT_Class:
                    constantPool[cpi]=new ClassInfo(is.readUnsignedShort());
                    break;
                case CONSTANT_Fieldref:
                    constantPool[cpi]=new FieldRef(is.readUnsignedShort(),is.readUnsignedShort());
                    break;
                case CONSTANT_Methodref:
                    final MethRef methRef = new MethRef(is.readUnsignedShort(),is.readUnsignedShort());
                    constantPool[cpi]=methRef;
                    methodRefs.add(methRef);
                    break;
                case CONSTANT_InterfaceMethodref://these all have the same structure,
                    constantPool[cpi]=new MethRef(is.readUnsignedShort(),is.readUnsignedShort());
                    break;
                case CONSTANT_String:
                    constantPool[cpi]=new StringInfo(is.readUnsignedShort());
                    break;
                case CONSTANT_Integer:
                    constantPool[cpi]=Integer.valueOf(is.readInt());
                    break;
                case CONSTANT_Float:
                    constantPool[cpi]=Float.valueOf(is.readFloat());
                    break;
                case CONSTANT_Long:
                    constantPool[cpi++]=Long.valueOf(is.readLong());
                    break;
                case CONSTANT_Double:
                    constantPool[cpi++]=Double.valueOf(is.readDouble());
                    // All 8-byte constants take up two entries in the constant_pool table of the class file.
                    // If a CONSTANT_Long_info or CONSTANT_Double_info structure is the item in the constant_pool
                    // table at index n, then the next usable item in the pool is located at index n+2.
                    // The constant_pool index n+1 must be valid but is considered unusable.2
                    //NS: the above is not at all clear to me, but looking at code on the web, it seems doing the above
                    //is what they are hinting at..
                    break;
                case CONSTANT_NameAndType:
                    constantPool[cpi]=new NameAndType(is.readUnsignedShort(),is.readUnsignedShort());
                    break;
                case CONSTANT_Utf8:
                    constantPool[cpi]=is.readUTF();
                    break;
                default:{
                    log.severe("unrecognized constantPool tagbyte="+tag+" (aborting)");
                    throw new BadClassFileException("unrecognized constantPool tagbyte="+tag);
                }
            }
        }

    }

    private MethodInfo readMethodInfo(DataInputStream is) throws IOException {
        MethodInfo fi = new MethodInfo();
        fi.accessFlags=is.readUnsignedShort();
        fi.nameIndex=is.readUnsignedShort();
        fi.descriptorIndex=is.readUnsignedShort();
        int attribscount=is.readUnsignedShort();
        fi.attributes=new AttributesInfo[attribscount];
        for(int i=0;i<attribscount;i++){
            fi.attributes[i]=readAttributesInfo(is);
        }
        return fi;
    }
    private AttributesInfo readAttributesInfo(DataInputStream is) throws IOException {
        AttributesInfo ai = new AttributesInfo();
        ai.nameIndex=is.readUnsignedShort();
        int len= is.readInt();
        ai.info=new byte[len];
        for(int i=0;i<len;i++){
            ai.info[i]=(byte) is.readUnsignedByte();
        }
        return ai;
    }
    private FieldInfo readFieldInfo(DataInputStream is) throws IOException {
        FieldInfo fi = new FieldInfo();
        fi.accessFlags=is.readUnsignedShort();
        fi.nameIndex=is.readUnsignedShort();
        fi.descriptorIndex=is.readUnsignedShort();
        int attribscount=is.readUnsignedShort();
        fi.attributes=new AttributesInfo[attribscount];
        for(int i=0;i<attribscount;i++){
            fi.attributes[i]=readAttributesInfo(is);
        }
        return fi;
    }



    //these represent datastructures which reference the constantpool for their values. We parse teh constantpool, after
    //which we can query these structures for their values and they will look them up from the constantpool
    private Object[] constantPool;

    public void setPrefixesToInclude(ArrayList<String> onlyPrefixList) {
        this.onlyPrefixList=onlyPrefixList;

    }

    private class ClassInfo{
        private int nameIndex;

        public ClassInfo(int nameIndex) {
            this.nameIndex = nameIndex;
        }

        public String getName(){
            return (String) constantPool[nameIndex];
        }

        public String toString() {
            return getClassName();
        }

        /**
         *
         * as per http://java.sun.com/docs/books/vmspec/2nd-edition/html/ClassFile.doc.html section 4.4.1
         * @return fully qualified classname encoded in the classinfoNameindex value. Or null if not a class (e.g. a primitive type)
         */
        public String getClassName() {
            String name=getName();
            //if it's an array, the rules are as for fields (e.g. [[I, [[Lclassname;
            //cf. http://java.sun.com/docs/books/vmspec/2nd-edition/html/ClassFile.doc.html section 4.4.1
            if(name.startsWith("[")){
                List<String> s= findClassnames(name);
                if(s.size()>1){
                    log.warning("classname="+name+" appears to contain more than 1 class: "+s+": ignoring all but the first");
                }
                if(!s.isEmpty()){
                    return s.get(0);
                }
                return null;
            }
            else return decodeInternalName(name);
        }

    }


    private abstract class FieldMethRef{
        private int classIndex;
        private int nameTypeIndex;

        public FieldMethRef(int classIndex, int nameTypeIndex) {
            this.classIndex = classIndex;
            this.nameTypeIndex = nameTypeIndex;
        }
        public ClassInfo getClassInfo(){
            return (ClassInfo) constantPool[classIndex];
        }
        public NameAndType getNameAndType(){
            return (NameAndType) constantPool[nameTypeIndex];
        }
         public String toString() {
            return "classinfo="+getClassInfo()+" nametype="+getNameAndType();
        }
    }
    private class FieldRef extends FieldMethRef{
        public FieldRef(int classIndex, int nameTypeIndex) {
            super(classIndex,nameTypeIndex);
        }
    }
    private class MethRef extends FieldMethRef{
        public MethRef(int classIndex, int nameTypeIndex) {
            super(classIndex,nameTypeIndex);
        }


    }
//    private class IntMethRef extends FieldMethRef{
//        public IntMethRef(int classIndex, int nameTypeIndex) {
//            super(classIndex,nameTypeIndex);
//        }
//    }

    private class NameAndType{
        private int nameIndex;
        private int descriptorIndex;

        public NameAndType(int nameIndex, int descriptorIndex) {
            this.nameIndex = nameIndex;
            this.descriptorIndex = descriptorIndex;
        }
        public String getName(){
            return (String) constantPool[nameIndex];
        }
        public String getType(){
            return (String) constantPool[descriptorIndex];
        }
        /**
         * extract any fully-qualififed class names from this method or field descriptor.
         * See
         * http://java.sun.com/docs/books/vmspec/2nd-edition/html/ClassFile.doc.html#1169
         * for details of how these are encoded
         * @return
         */
        public List<String> getReferencedClassnames() {
            String encodedDescriptor=getType();
            return findClassnames(encodedDescriptor);
        }

        public String toString() {
            return "name="+getName()+" type="+getType();
        }
    }
    private class StringInfo{
        private int stringIndex;

        public StringInfo(int stringIndex) {
            this.stringIndex = stringIndex;
        }
        public String getValue(){
            return (String) constantPool[stringIndex];
        }
    }
     private class AttributesInfo{
        private int nameIndex;
        private byte[] info;
    }
    private class FieldInfo{
        private int accessFlags,nameIndex,descriptorIndex;
        private AttributesInfo[] attributes;
        public String getDescriptor(){
            return (String) constantPool[descriptorIndex];
        }
    }
    private class MethodInfo{
        private int accessFlags,nameIndex,descriptorIndex;
        private AttributesInfo[] attributes;
        public String getDescriptor(){
            return (String) constantPool[descriptorIndex];
        }
    }
}
